/** 
 * @file  RegistryBundle.cpp
 *
 * @brief Implementation of CRegKeyEx C++ wrapper class for reading Windows registry
 */
// RCS ID line follows -- this is updated by CVS
// $Id: RegistryBundle.cpp,v 1.5 2006/03/04 17:54:58 elsapo Exp $

#include "stdafx.h"	
#include <assert.h>
#include "RegistryBundle.h"

#define MIN_BUFFER_DELTA 128

bool RegVal::Clear()
{
  //bool b = dwType == REG_SZ || dwType == REG_EXPAND_SZ
  //  || dwType == REG_LINK || dwType == REG_MULTI_SZ
  //  || dwType == REG_BINARY;
  bool b = m_dwAllocSize;
  if (b)
  {
    //free (pValData->pbyteData);
    delete[] reinterpret_cast<LPBYTE>(val.LowPart);
    val.LowPart = NULL;
    m_dwAllocSize = 0;
  }
  dwSize = 0;
  dwType = REG_NONE;
  return b;
}

RegVal::RegVal(const RegVal &src)
{
  val.QuadPart = 0;
  m_dwAllocSize = src.m_dwAllocSize;
  dwType = src.dwType;
  dwSize = src.dwSize;
  if (m_dwAllocSize)
  {
    val.LowPart = reinterpret_cast<DWORD>(new BYTE[m_dwAllocSize]);
    memcpy_s(reinterpret_cast<LPBYTE>(val.LowPart), m_dwAllocSize, reinterpret_cast<LPBYTE>(src.val.LowPart), dwSize);
  }
  else
    val = src.val;
}

DWORD RegVal::AllocateMemory(DWORD size, bool bRealloc)
{
  if (bRealloc && (m_dwAllocSize - size > MIN_BUFFER_DELTA) || (m_dwAllocSize < size) || !size)
    Clear();
  if (size)
  {
    if (!m_dwAllocSize)
    {
      val.LowPart = reinterpret_cast<DWORD>(new BYTE[size]);
      m_dwAllocSize = size;
    }
    dwType = REG_BINARY;
    dwSize = size;
  }
  return m_dwAllocSize;
}

bool RegVal::Initialize(DWORD type, DWORD size)
{
  if (type == REG_NONE)
  {
    Clear();
    return true;
  }
  if (((type == REG_DWORD_LITTLE_ENDIAN) || (type == REG_DWORD_BIG_ENDIAN)) && (size == sizeof(DWORD)))
  {
    DwordValue(0, type == REG_DWORD_BIG_ENDIAN);
    return true;
  }
  if ((type == REG_QWORD_LITTLE_ENDIAN) && (size == sizeof(ULONGLONG)))
  {
    QwordValue(0);
    return true;
  }
  if ((type == REG_SZ || type == REG_EXPAND_SZ || type == REG_MULTI_SZ || type == REG_BINARY) && (size > 0))
  {
    bool b = AllocateMemory(size) >= size;
    if (b)
      dwType = type;
    return b;
  }
  return false;
}

DWORD RegVal::StringValue(LPTSTR lpBuffer, DWORD dwBufferCharLength) const
{
  if (IsString())
  {
    DWORD dwLen = dwBufferCharLength * sizeof(TCHAR);
    DWORD dwCharLen = (dwSize / sizeof(TCHAR)) + (dwSize % sizeof(TCHAR));
    if (lpBuffer && (dwSize <= dwLen))
    {
      memcpy_s(lpBuffer, dwLen, reinterpret_cast<LPBYTE>(val.LowPart), dwSize);
      if (dwCharLen < dwBufferCharLength)
        lpBuffer[dwCharLen] = 0;
    }
    return dwCharLen;
  }
  return 0;
}

DWORD RegVal::BinaryValue(LPBYTE lpBuffer, DWORD dwBufferSize) const
{
  if (IsBinary())
  {
    if (lpBuffer && (dwSize <= dwBufferSize))
    {
      memcpy_s(lpBuffer, dwBufferSize, reinterpret_cast<LPBYTE>(val.LowPart), dwSize);
      if (dwSize < dwBufferSize)
        lpBuffer[dwSize] = 0;
    }
    return dwSize;
  }
  return 0;
}

DWORD RegVal::StringValue(LPTSTR &lpBuffer) const
{
  DWORD dwSz = StringValue(NULL, 0);
  if (dwSz)
  {
    lpBuffer = new TCHAR[++dwSz];
    dwSz = StringValue(lpBuffer, dwSz);
  }
  return dwSz;
}

DWORD RegVal::BinaryValue(LPBYTE &lpBuffer) const
{
  DWORD dwSz = BinaryValue(NULL, 0);
  if (dwSz)
  {
    lpBuffer = new BYTE[++dwSz];
    dwSz = BinaryValue(lpBuffer, dwSz);
  }
  return dwSz;
}

bool RegVal::StringValue(LPCTSTR lpBuffer, DWORD dwBufferCharLength, bool bRealloc, DWORD dwStringType /*= REG_SZ*/)
{
  if (lpBuffer && dwBufferCharLength && (dwStringType == REG_SZ || dwStringType == REG_EXPAND_SZ || dwStringType == REG_MULTI_SZ))
  {
    DWORD dwLen = dwBufferCharLength * sizeof(TCHAR);
    /*if (bRealloc || (m_dwAllocSize < dwLen))
      Clear();
    if (!m_dwAllocSize)
    {
      val.LowPart = reinterpret_cast<DWORD>(new BYTE[dwLen]);
      m_dwAllocSize = dwLen;
    }*/
    if (AllocateMemory(dwLen, bRealloc) >= dwLen)
    {
      //dwSize = dwLen;
      dwType = dwStringType;
      memcpy_s(reinterpret_cast<LPBYTE>(val.LowPart), m_dwAllocSize, lpBuffer, dwLen);
      return true;
    }
  }
  return false;
}

bool RegVal::BinaryValue(const BYTE *lpBuffer, DWORD dwBufferSize, bool bRealloc)
{
  if (lpBuffer && dwBufferSize)
  {
    /*if (bRealloc || (m_dwAllocSize < dwBufferSize))
      Clear();
    if (!m_dwAllocSize)
    {
      val.LowPart = reinterpret_cast<DWORD>(new BYTE[dwBufferSize]);
      m_dwAllocSize = dwBufferSize;
    }*/
    if (AllocateMemory(dwBufferSize, bRealloc) >= dwBufferSize)
    {
      //dwSize = dwBufferSize;
      //dwType = REG_BINARY;
      memcpy_s(reinterpret_cast<LPBYTE>(val.LowPart), m_dwAllocSize, lpBuffer, dwBufferSize);
      return true;
    }
  }
  return false;
}
//----------------------------------------------------------------------------
/*
CRegKeyEx::CRegKeyEx()
{
	m_hKey = NULL;
  m_sPath[0] = 0;
}

CRegKeyEx::~CRegKeyEx()
{
	Close();
}
*/
void CRegKeyEx::Close(bool bOnlyCurrent)
{
	if (m_hKey) 
	{
		m_OperationStatus = RegCloseKey(m_hKey);
		m_hKey = NULL;
    m_sPath[0] = 0;
	}
  if (m_bOwned && !bOnlyCurrent && m_hRootKey)
  {
		RegCloseKey(m_hRootKey);
		m_hRootKey = NULL;
    m_sMachine[0] = 0;
    m_bOwned = false;
	}
}

LSTATUS CRegKeyEx::OpenRootKey(HKEY hKeyRoot, LPCTSTR lpRemoteMachine)
{
	Close();
  LSTATUS lRes = ERROR_SUCCESS;
  if (lpRemoteMachine && *lpRemoteMachine)
  {
    _tcscpy_s(m_sMachine, lpRemoteMachine);
    lRes = RegConnectRegistry(lpRemoteMachine, hKeyRoot, &m_hRootKey);
    m_bOwned = (lRes == ERROR_SUCCESS);
  }
  else
    m_hRootKey = hKeyRoot;
  return lRes;
}

LSTATUS CRegKeyEx::Open(HKEY hKeyRoot, LPCTSTR pszPath, LPCTSTR lpRemoteMachine)
{
	return OpenWithAccess(hKeyRoot, pszPath, KEY_ALL_ACCESS, lpRemoteMachine);
}

LSTATUS CRegKeyEx::OpenWithAccess(HKEY hKeyRoot, LPCTSTR pszPath, REGSAM regsam, LPCTSTR lpRemoteMachine)
{
  m_OperationStatus = OpenRootKey(hKeyRoot, lpRemoteMachine);
  if (m_OperationStatus == ERROR_SUCCESS)
  {
    if (pszPath && *pszPath)
	    _tcscpy_s(m_sPath,pszPath);
	  m_OperationStatus = RegCreateKeyEx(m_hRootKey, m_sPath, 0L, NULL,
		  REG_OPTION_NON_VOLATILE, regsam, NULL, 
		  &m_hKey, NULL);
  }
  return m_OperationStatus;
}

LSTATUS CRegKeyEx::OpenNoCreateWithAccess(HKEY hKeyRoot, LPCTSTR pszPath, REGSAM regsam, LPCTSTR lpRemoteMachine)
{
  m_OperationStatus = OpenRootKey(hKeyRoot, lpRemoteMachine);
  if (m_OperationStatus == ERROR_SUCCESS)
  {
    if (pszPath && *pszPath)
	    _tcscpy_s(m_sPath,pszPath);
	  m_OperationStatus = RegOpenKeyEx(m_hRootKey, m_sPath, 0L, regsam, &m_hKey);
  }
  return m_OperationStatus;
}

// Open of registry key under HKEY_LOCAL_MACHINE
bool
CRegKeyEx::OpenRegMachine(LPCTSTR key, REGSAM regsam, LPCTSTR lpRemoteMachine, bool bCreate)
{
  return ((bCreate) ? OpenWithAccess(HKEY_LOCAL_MACHINE, key, regsam, lpRemoteMachine) : OpenNoCreateWithAccess(HKEY_LOCAL_MACHINE, key, regsam, lpRemoteMachine)) == ERROR_SUCCESS;
}

// Open of registry key under HKEY_CURRENT_USER
bool
CRegKeyEx::OpenRegUser(LPCTSTR key, REGSAM regsam, LPCTSTR lpRemoteMachine, bool bCreate)
{
  if (lpRemoteMachine && *lpRemoteMachine)
    return ((bCreate) ? OpenWithAccess(HKEY_CURRENT_USER, key, regsam, lpRemoteMachine) : OpenNoCreateWithAccess(HKEY_CURRENT_USER, key, regsam, lpRemoteMachine)) == ERROR_SUCCESS;
  Close();
  if ((m_OperationStatus = RegOpenCurrentUser(regsam, &m_hRootKey)) == ERROR_SUCCESS)
  {
    LSTATUS operationStatus = (bCreate) ? OpenWithAccess(m_hRootKey, key, regsam) : OpenNoCreateWithAccess(m_hRootKey, key, regsam);
    m_bOwned = true;
    return operationStatus == ERROR_SUCCESS;
  }
  return false;
}

LSTATUS CRegKeyEx::WriteValue(LPCTSTR pszSubKey, LPCTSTR pszValueName, DWORD dwType, const BYTE* lpData, DWORD cbData)
{
	//assert(m_hKey);
	//assert(pszKey);

  //return (pszSubKey && *pszSubKey) ? RegSetKeyValue(m_hKey, pszSubKey, pszValueName, dwType, lpData, cbData) : RegSetValueEx(m_hKey, pszValueName, 0L, dwType, lpData, cbData);
  return m_OperationStatus = RegSetKeyValue(m_hKey, pszSubKey, pszValueName, dwType, lpData, cbData);
}

LSTATUS CRegKeyEx::ReadValue(LPCTSTR pszSubKey, LPCTSTR pszValueName, LPDWORD lpdwType, LPBYTE lpData, DWORD &cbData) const
{
	//assert(m_hKey);
	//assert(pszKey);

	//return RegQueryValueEx(m_hKey, pszKey, NULL, lpdwType, lpData, &cbData);
  const_cast<CRegKeyEx*>(this)->m_OperationStatus = RegGetValue(m_hKey, pszSubKey, pszValueName, RRF_RT_ANY, lpdwType, lpData, &cbData);
  return m_OperationStatus;
}

LSTATUS CRegKeyEx::WriteValue(LPCTSTR pszSubKey, LPCTSTR pszValueName, const RegVal *rvData)
{
  return (rvData) ? WriteValue(pszSubKey, pszValueName, rvData->DataType(), rvData->RawValue(), rvData->DataSize()) : ERROR_INVALID_PARAMETER;
}

LSTATUS CRegKeyEx::ReadValue(LPCTSTR pszSubKey, LPCTSTR pszValueName, RegVal *rvData) const
{
  LSTATUS lRes = ERROR_INVALID_PARAMETER;
	if (rvData)
  {
    DWORD dwTp = REG_NONE;
    DWORD dwSz = 0;
    lRes = ReadValue(pszSubKey, pszValueName, &dwTp, NULL, dwSz);
    if (lRes == ERROR_SUCCESS)
      lRes = (rvData->Initialize(dwTp, dwSz)) ? ReadValue(pszSubKey, pszValueName, &dwTp, (LPBYTE)rvData->RawValue(), dwSz) : ERROR_INVALID_DATA;
  }
  return lRes;
}
/*
LONG CRegKeyEx::WriteFloat (LPCTSTR pszKey, float fVal)
{
	assert(m_hKey);
	assert(pszKey);
	TCHAR pszData[100];
	_stprintf_s(pszData,_T("%f"),fVal);
	return RegSetValueEx (m_hKey, pszKey, 0L, REG_SZ,
		(const LPBYTE) pszData, (_tcslen(pszData)+ 1)*sizeof(TCHAR) );
}
*/
LSTATUS CRegKeyEx::WriteString(LPCTSTR pszSubKey, LPCTSTR pszValueName, LPCTSTR pszData)
{
	//assert(m_hKey);
	//assert(pszKey);
	//assert(pszData);
//	assert(AfxIsValidAddress(pszData, strlen(pszData), FALSE));

	return WriteValue(pszSubKey, pszValueName, REG_SZ, (const BYTE*) pszData, pszData ? ((_tcslen(pszData)+1)*sizeof(TCHAR)) : 0);
}

LSTATUS CRegKeyEx::WriteMultiString(LPCTSTR pszSubKey, LPCTSTR pszValueName, LPCTSTR pszData, DWORD dwCharLen)
{
	//assert(m_hKey);
	//assert(pszKey);
	//assert(pszData);
//	assert(AfxIsValidAddress(pszData, strlen(pszData), FALSE));

  return WriteValue(pszSubKey, pszValueName, REG_MULTI_SZ, (const BYTE*) pszData, (pszData) ? dwCharLen*sizeof(TCHAR) : 0);
}

/*template<class T>
T CRegKeyEx::ReadNumeric(LPCTSTR pszSubKey, LPCTSTR pszValueName, T defval, LPDWORD lpdwType) const
{
	//assert(m_hKey);
	//assert(pszKey);

	T dwDest;
	DWORD dwSize = sizeof(dwDest);

	LSTATUS operationStatus = ReadValue(pszSubKey, pszValueName, lpdwType, (LPBYTE) &dwDest, dwSize);

	if (operationStatus == ERROR_SUCCESS)
		return dwDest;
	return defval;
}*/
/*
LONG CRegKeyEx::ReadLong (LPCTSTR pszKey, LONG defval)
{
	return (LONG)ReadDword(pszKey, (DWORD)defval);
}

UINT CRegKeyEx::ReadUint (LPCTSTR pszKey, UINT defval)
{
	return (UINT)ReadDword(pszKey, (DWORD)defval);
}

int CRegKeyEx::ReadInt (LPCTSTR pszKey, int defval)
{
	return (int)ReadDword(pszKey, (DWORD)defval);
}

short int CRegKeyEx::ReadShort (LPCTSTR pszKey, short int defval)
{
	return (short int)ReadDword(pszKey, (DWORD)defval);
}

BYTE CRegKeyEx::ReadByte (LPCTSTR pszKey, BYTE defval)
{
	return (BYTE)ReadDword(pszKey, (DWORD)defval);
}

float CRegKeyEx::ReadFloat (LPCTSTR pszKey, float defval)
{
	assert(m_hKey);
	assert(pszKey);

	DWORD dwType;
	TCHAR  string[100];
	DWORD dwSize = sizeof string;

	LONG lReturn = RegQueryValueEx (m_hKey, pszKey, NULL,
		&dwType, (LPBYTE) string, &dwSize);

	if (lReturn == ERROR_SUCCESS)
		return (float)_tcstod(string, NULL);
	return defval;
}

BOOL CRegKeyEx::ReadBool (LPCTSTR pszKey, BOOL defval)
{
  return ReadDword(pszKey, (DWORD)defval) != 0;
}
*/
CString CRegKeyEx::ReadString(LPCTSTR pszSubKey, LPCTSTR pszValueName, LPCTSTR defval) const
{
	//assert(m_hKey);
	//assert(pszKey);

  CString string(defval);
	DWORD dwType;
	DWORD dwSize = 0;

	LSTATUS operationStatus = ReadValue(pszSubKey, pszValueName, &dwType, NULL, dwSize);

	if ((operationStatus == ERROR_SUCCESS) && ((dwType == REG_SZ) || (dwType == REG_EXPAND_SZ)))
  {
    dwSize /= sizeof(TCHAR);
		LPTSTR buff = new TCHAR[++dwSize];
    dwSize *= sizeof(TCHAR);
    operationStatus = ReadValue(pszSubKey, pszValueName, &dwType, (LPBYTE)buff, dwSize);
    if (operationStatus == ERROR_SUCCESS)
      string = buff;
    delete[] buff;
  }
	return string;
}

DWORD CRegKeyEx::ReadChars(LPCTSTR pszSubKey, LPCTSTR pszValueName, LPTSTR pData, DWORD dwCharLength) const
{
	//assert(m_hKey);
	//assert(pszKey);
  DWORD dwRes = 0;
  if (pData && dwCharLength)
  {
    DWORD dwType = REG_NONE;
    DWORD dwSize = 0;
    LSTATUS operationStatus = ReadValue(pszSubKey, pszValueName, &dwType, NULL, dwSize);
    if ((operationStatus == ERROR_SUCCESS) && ((dwType == REG_SZ) || (dwType == REG_EXPAND_SZ) || (dwType == REG_MULTI_SZ)))
    {
      dwSize /= sizeof(TCHAR);
      if (dwSize <= dwCharLength)
      {
        dwSize = dwCharLength * sizeof(TCHAR);
        operationStatus = ReadValue(pszSubKey, pszValueName, &dwType, (LPBYTE)pData, dwSize);
        if (operationStatus == ERROR_SUCCESS)
          dwRes = dwSize / sizeof(TCHAR);
      }
    }
  }
  return dwRes;
}

LSTATUS CRegKeyEx::EnumValue(DWORD dwIndex,LPTSTR lpValueName,LPDWORD lpcValueName,LPDWORD lpType,LPBYTE lpData,LPDWORD lpcbData) const
{
  const_cast<CRegKeyEx*>(this)->m_OperationStatus = RegEnumValue(m_hKey, dwIndex, lpValueName, lpcValueName, NULL, lpType, lpData, lpcbData);
  return m_OperationStatus;
}

LSTATUS CRegKeyEx::EnumValue(DWORD dwIndex,LPTSTR lpValueName,LPDWORD lpcValueName, RegVal *rvData) const
{
  LSTATUS operationStatus = ERROR_INVALID_PARAMETER;
	if (rvData)
  {
    DWORD dwTp = REG_NONE;
    DWORD dwSz = 0;
    operationStatus = EnumValue(dwIndex, lpValueName, lpcValueName, &dwTp, NULL, &dwSz);
    if (operationStatus == ERROR_SUCCESS)
      operationStatus = (rvData->Initialize(dwTp, dwSz)) ? EnumValue(dwIndex, lpValueName, lpcValueName, &dwTp, (LPBYTE)rvData->RawValue(), &dwSz) : ERROR_INVALID_DATA;
  }
  return operationStatus;
}

LSTATUS CRegKeyEx::DeleteKey()
{
  Close(true);
  LSTATUS lRes = m_OperationStatus = RegDeleteKey(m_hRootKey, m_sPath);
  if (lRes == ERROR_SUCCESS)
    Close();
  return lRes;
}

LSTATUS CRegKeyEx::DeleteTree()
{
  Close(true);
  LSTATUS lRes = m_OperationStatus = RegDeleteTree(m_hRootKey, m_sPath);
  if (lRes == ERROR_SUCCESS)
    Close();
  return lRes;
}

LSTATUS CRegKeyEx::DeleteValue(LPCTSTR pszValueName)
{
  return m_OperationStatus = RegDeleteValue(m_hKey, pszValueName);
}

LSTATUS CRegKeyEx::DeleteKeyValue(LPCTSTR pszSubKey, LPCTSTR pszValueName)
{
  return m_OperationStatus = RegDeleteKeyValue(m_hKey, pszSubKey, pszValueName);
}

LSTATUS CRegKeyEx::QueryInfoKey(LPTSTR lpClass,LPDWORD lpcClass,LPDWORD lpcSubKeys,LPDWORD lpcMaxSubKeyLen,LPDWORD lpcMaxClassLen,LPDWORD lpcValues,LPDWORD lpcMaxValueNameLen,LPDWORD lpcMaxValueLen,LPDWORD lpcbSecurityDescriptor,PFILETIME lpftLastWriteTime) const
{
  const_cast<CRegKeyEx*>(this)->m_OperationStatus = RegQueryInfoKey(m_hKey, lpClass, lpcClass, NULL, lpcSubKeys, lpcMaxSubKeyLen, lpcMaxClassLen, lpcValues, lpcMaxValueNameLen, lpcMaxValueLen, lpcbSecurityDescriptor, lpftLastWriteTime);
  return m_OperationStatus;
}