/** 
 * @file  sbuffer.h
 *
 * @brief Declaration file for sbuffer class
 *
 * @date  Created: 2003-10-09 (Perry Rapp)
 */
// RCS ID line follows -- this is updated by CVS
// $Id: sbuffer.h 1949 2004-11-06 19:59:50Z puddle $

#ifndef sbuffer_h_included
#define sbuffer_h_included

#define BUFFER_SIZE_DEFAULT 8192
#define BUFFER_SIZE_DELTA 1024
#define MIN_BUFFER_DELTA BUFFER_SIZE_DELTA
/**
 * @brief Primitive string that grows with exponential allocation
 */
class sbuffer
{
	LPTSTR m_buff;
	UINT m_len;
	UINT m_size;
  void Destroy()
  {
    delete[] m_buff;
	  m_buff = NULL;
    m_size = 0;
    m_len = 0;
  }
public:
	sbuffer(UINT init = 0)
	{
    m_size = (init) ? init : BUFFER_SIZE_DEFAULT;
		m_buff = new TCHAR[m_size];
		m_len = 0;
	}
	~sbuffer()
	{
		Destroy();
	}
	void Append(LPCTSTR sz)
	{
    if (sz)
    {
      int len = _tcslen(sz);
      if (len > 0)
      {
        int nlen = m_len + len;
        if (nlen >= m_size)
        {
          m_size = nlen + BUFFER_SIZE_DELTA + 1;
          LPTSTR nbuf = new TCHAR[m_size];
          nbuf[0] = 0;
          if (m_buff)
          {
            _tcscpy_s(nbuf, m_size, m_buff);
            delete[] m_buff;
          }
          m_buff = nbuf;
        }
        _tcscat_s(m_buff, m_size, sz);
        m_len = nlen;
      }
    }
		//CString str = sz;
		//Append(str);
	}
	/*void Append(const CString & s)
		{
			//if (s.GetLength() + m_len > m_size-1)
			//{
			//	do
			//	{
			//		m_size = m_size * 2;
			//	}
			//	while (s.GetLength() + m_len > m_size-1);
			//	LPTSTR newbuff = new TCHAR[m_size];
			//	_tcscpy_s(newbuff, m_size, m_buff);
			//	delete[] m_buff;
			//	m_buff = newbuff;
			//}
			//_tcscat_s(m_buff, m_size, s);
			//m_len += s.GetLength();
      Append(s);
		}*/
	void Clear()
  {
    if (m_buff)
    {
      m_buff[0] = 0;
      m_len = 0;
    }
  }
	void Set(LPCTSTR sztext)
	{
    int len = (sztext) ? _tcslen(sztext) : 0;
    if (m_size - len > MIN_BUFFER_DELTA)
      Destroy();
    Clear();
    if (len > 0)
		  Append(sztext);
	}
	bool StartsWithInsensitive(LPCTSTR prefix)
	{
    return (m_buff) ? !_tcsnicmp(m_buff, prefix, _tcslen(prefix)) : false;
	}
	CString Left(int len)
	{
		CString s;
    if (m_buff)
    {
		  LPTSTR sz = s.GetBuffer(len+1);
		  _tcsncpy_s(sz, len+1, m_buff, len);
		  //sz[len] = 0;
		  s.ReleaseBuffer();
    }
		return s;
	}
	// get data length
	int GetLength() const { return m_len; }

	// TRUE if zero length
	bool IsEmpty() const { return !m_len; }
	
	LPCTSTR GetData() const { return m_buff; }
  operator LPCTSTR() const { return m_buff; }
  LPCTSTR operator=(LPCTSTR sztext) {Set(sztext); return m_buff;}
  LPCTSTR operator=(const CString &stext) {Set(stext); return m_buff;}
};

#endif // sbuffer_h_included
