///////////////////////////////////////////////////////////////////////////
//  File:    syntaxbase.cpp
//
////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "syntaxbase.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

typedef int (*S_CMP_1)(LPCTSTR, LPCTSTR);
typedef int (*S_NCMP_1)(LPCTSTR, LPCTSTR, size_t);
  
BOOL LangSyntaxBase::IsXKeyword(LPTSTR apszKeywords[], LPCTSTR pszChars, bool caseInsensitive, int nLength)
{
  S_CMP_1 s1 = _tcscmp;
  S_NCMP_1 sn1 = _tcsncmp;
  if (caseInsensitive)
  {
    s1 = _tcsicmp;
    sn1 = _tcsnicmp;
  }
  if (nLength > 0)
  {
    for (int L = 0; apszKeywords[L]; ++L)
      if (!sn1(apszKeywords[L], pszChars, nLength) && !apszKeywords[L][nLength])
        return TRUE;
  }
  else
  {
    for (int L = 0; apszKeywords[L]; ++L)
      if (!s1(apszKeywords[L], pszChars))
        return TRUE;
  }
  return FALSE;
}

BOOL LangSyntaxBase::IsNumber(LPCTSTR pszChars, int nLength)
{
  if (nLength <= 0)
    nLength = lstrlen(pszChars);
  if (nLength > 0)
  {
    int i = 0;
    if (pszChars[0] == '-')
      i = 1;
    if (nLength > i)
    {
      for (; i < nLength; ++i)
        if (!_istdigit(pszChars[i]))
          return FALSE;
      return TRUE;
    }
  }
  return FALSE;
}

BOOL LangSyntaxBase::IsNumber2(LPCTSTR pszChars, LPCTSTR sHexPrefix, LPCTSTR sLongSuffix, bool bRealAllowed, int nLength)
{
  if (nLength <= 0)
    nLength = lstrlen(pszChars);
  if (nLength > 0)
  {
    bool bFl = false;
    int nLL = (sLongSuffix) ? lstrlen(sLongSuffix) : 0;
    int nHL = (sHexPrefix) ? lstrlen(sHexPrefix) : 0;
    int I = nHL;
    if ((nHL > 0) && (nLength > nHL) && !_tcsnicmp(pszChars, sHexPrefix, nHL))
    {
      for (; I < nLength; ++I)
        if (!(_istdigit(pszChars[I]) || (pszChars[I] >= 'A' && pszChars[I] <= 'F') ||
              (pszChars[I] >= 'a' && pszChars[I] <= 'f')))
        {
          bFl = true;
          break;
        }
      if (bFl)
        return (nLL > 0) && MatchSuffix(&pszChars[I], sLongSuffix, false, nLength - I);
      return TRUE;
    }
    if (pszChars[0] == '-')
    {
      I = 2;
      if (!((nLength > 1) && _istdigit(pszChars[1])))
        return FALSE;
    }
    else
    {
      I = 1;
      if (!_istdigit(pszChars[0]))
        return FALSE;
    }
    bool bR = false;
    for (; I < nLength; ++I)
    {
      if (!_istdigit(pszChars[I]))
      {
        if (bRealAllowed && (pszChars[I] == '+' ||
            pszChars[I] == '-' || pszChars[I] == '.' || pszChars[I] == 'e' ||
            pszChars[I] == 'E'))
          bR = true;
        else
        {
          bFl = true;
          break;
        }
      }
    }
    if (bFl)
      return (nLL > 0) && MatchSuffix(&pszChars[I], sLongSuffix, bR, nLength - I);
    return TRUE;
  }
  return FALSE;
}

bool LangSyntaxBase::MatchSuffix(LPCTSTR pszChars, LPCTSTR sLongSuffix, bool bReal, int nLength)
{
  if (nLength <= 0)
    nLength = lstrlen(pszChars);
  if ((nLength > 0) && sLongSuffix && *sLongSuffix)
  {
    TCHAR buff[MAX_PATH];
    _tcsncpy_s(buff, sLongSuffix, _TRUNCATE);
    LPCTSTR delim = _T(" ");
    LPTSTR ttt = NULL;
    LPTSTR tok = _tcstok_s(buff, delim, &ttt);
    while (tok)
    {
      int nn = lstrlen(tok);
      bool bR1 = (nn > 1) && (tok[0] == REAL_SUFFIX_MARK);
      if (bR1)
      {
        ++tok;
        --nn;
      }
      if ((bR1 == bReal) && (nn == nLength) && !_tcsnicmp(pszChars, tok, nn))
        return true;
      tok = _tcstok_s(NULL, delim, &ttt);
    }
  }
  return false;
}