///////////////////////////////////////////////////////////////////////////
//  File:    vb.cpp
//  Version: 1.1.0.4
//  Updated: 19-Jul-1998
//
//  Copyright:  Ferdinand Prantl, portions by Stcherbatchenko Andrei
//  E-mail:     prantl@ff.cuni.cz
//
//  Visual VB syntax highlighing definition
//
//  You are free to use or modify this code to the following restrictions:
//  - Acknowledge me somewhere in your about box, simple "Parts of code by.."
//  will be enough. If you can't (or don't want to), contact me personally.
//  - LEAVE THIS HEADER INTACT
////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "../ccrystaltextview.h"
#include "../ccrystaltextbuffer.h"
#include "syntaxbase.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

static LPTSTR s_apszVBKeywordList[] =
{
  _T("AddHandler"),
  _T("AddressOf"),
  _T("Alias"),
  _T("And"),
  _T("AndAlso"),
  _T("As"),
  _T("Boolean"),
  _T("ByRef"),
  _T("Byte"),
  _T("ByVal"),
  _T("Call"),
  _T("Case"),
  _T("Catch"),
  _T("CBool"),
  _T("CByte"),
  _T("CChar"),
  _T("CDate"),
  _T("CDec"),
  _T("CDbl"),
  _T("Char"),
  _T("CInt"),
  _T("Class"),
  _T("CLng"),
  _T("CObj"),
  _T("Const"),
  _T("Continue"),
  _T("CSByte"),
  _T("CShort"),
  _T("CSng"),
  _T("CStr"),
  _T("CType"),
  _T("CUInt"),
  _T("CULng"),
  _T("CUShort"),
  _T("Date"),
  _T("Decimal"),
  _T("Declare"),
  _T("Default"),
  _T("Delegate"),
  _T("Dim"),
  _T("DirectCast"),
  _T("Do"),
  _T("Double"),
  _T("Each"),
  _T("Else"),
  _T("ElseIf"),
  _T("End"),
  _T("EndIf"),
  _T("Enum"),
  _T("Erase"),
  _T("Error"),
  _T("Event"),
  _T("Exit"),
  _T("False"),
  _T("Finally"),
  _T("For"),
  _T("Friend"),
  _T("Function"),
  _T("Get"),
  _T("GetType"),
  _T("Global"),
  _T("GoSub"),
  _T("GoTo"),
  _T("Handles"),
  _T("If"),
  _T("Implements"),
  _T("Imports"),
  _T("In"),
  _T("Inherits"),
  _T("Integer"),
  _T("Interface"),
  _T("Is"),
  _T("IsNot"),
  _T("Let"),
  _T("Lib"),
  _T("Like"),
  _T("Long"),
  _T("Loop"),
  _T("Me"),
  _T("Mod"),
  _T("Module"),
  _T("MustInherit"),
  _T("MustOverride"),
  _T("MyBase"),
  _T("MyClass"),
  _T("Namespace"),
  _T("Narrowing"),
  _T("New"),
  _T("Next"),
  _T("Not"),
  _T("Nothing"),
  _T("NotInheritable"),
  _T("NotOverridable"),
  _T("Object"),
  _T("Of"),
  _T("On"),
  _T("Operator"),
  _T("Option"),
  _T("Optional"),
  _T("Or"),
  _T("OrElse"),
  _T("Overloads"),
  _T("Overridable"),
  _T("Overrides"),
  _T("ParamArray"),
  _T("Partial"),
  _T("Private"),
  _T("Property"),
  _T("Protected"),
  _T("Public"),
  _T("RaiseEvent"),
  _T("ReadOnly"),
  _T("ReDim"),
  _T("REM"),
  _T("RemoveHandler"),
  _T("Resume"),
  _T("Return"),
  _T("SByte"),
  _T("Select"),
  _T("Set"),
  _T("Shadows"),
  _T("Shared"),
  _T("Short"),
  _T("Single"),
  _T("Static"),
  _T("Step"),
  _T("Stop"),
  _T("String"),
  _T("Structure"),
  _T("Sub"),
  _T("SyncLock"),
  _T("Then"),
  _T("Throw"),
  _T("To"),
  _T("True"),
  _T("Try"),
  _T("TryCast"),
  _T("TypeOf"),
  _T("Variant"),
  _T("Wend"),
  _T("UInteger"),
  _T("ULong"),
  _T("UShort"),
  _T("Using"),
  _T("When"),
  _T("While"),
  _T("Widening"),
  _T("With"),
  _T("WithEvents"),
  _T("WriteOnly"),
  _T("Xor"),
  _T("#Const"),
  _T("#Else"),
  _T("#ElseIf"),
  _T("#End"),
  _T("#If"),
  NULL
};
  
class LangVB : LangSyntaxBase
{
public:
  static BOOL IsVBNumber(LPCTSTR pszChars, int nLength);
  static BOOL IsVBKeyword(LPCTSTR pszChars, int nLength);
};

BOOL LangVB::IsVBKeyword (LPCTSTR pszChars, int nLength)
{
  return IsXKeyword (s_apszVBKeywordList, pszChars, true, nLength);
}

BOOL LangVB::IsVBNumber (LPCTSTR pszChars, int nLength)
{
  return IsNumber2(pszChars, _T("&H"), _T("L"), true, nLength);
}

DWORD CCrystalTextView::
ParseLineVB (DWORD dwCookie, int nLineIndex, TEXTBLOCK * pBuf, int &nActualItems)
{
  int nLength = GetLineLength (nLineIndex);
  if (nLength == 0)
    return dwCookie & COOKIE_EXT_COMMENT;

  LPCTSTR pszChars = GetLineChars (nLineIndex);
  BOOL bFirstChar = (dwCookie & ~COOKIE_EXT_COMMENT) == 0;
  BOOL bRedefineBlock = TRUE;
  BOOL bDecIndex = FALSE;
  int nIdentBegin = -1;
  int nPrevI = -1;
  int I=0;
  for (I = 0;; nPrevI = I, I = ::CharNext(pszChars+I) - pszChars)
    {
      if (bRedefineBlock)
        {
          int nPos = I;
          if (bDecIndex)
            nPos = nPrevI;
          if (dwCookie & (COOKIE_COMMENT | COOKIE_EXT_COMMENT))
            {
              DEFINE_BLOCK (nPos, COLORINDEX_COMMENT);
            }
          else if (dwCookie & (COOKIE_CHAR | COOKIE_STRING))
            {
              DEFINE_BLOCK (nPos, COLORINDEX_STRING);
            }
          else
            {
              if (xisalnum (pszChars[nPos]) || pszChars[nPos] == '.' && nPos > 0 && (!xisalpha (*::CharPrev(pszChars, pszChars + nPos)) && !xisalpha (*::CharNext(pszChars + nPos))))
                {
                  DEFINE_BLOCK (nPos, COLORINDEX_NORMALTEXT);
                }
              else
                {
                  DEFINE_BLOCK (nPos, COLORINDEX_OPERATOR);
                  bRedefineBlock = TRUE;
                  bDecIndex = TRUE;
                  goto out;
                }
            }
          bRedefineBlock = FALSE;
          bDecIndex = FALSE;
        }
out:

      // Can be bigger than length if there is binary data
      // See bug #1474782 Crash when comparing SQL with with binary data
      if (I >= nLength)
        break;

      if (dwCookie & COOKIE_COMMENT)
        {
          DEFINE_BLOCK (I, COLORINDEX_COMMENT);
          dwCookie |= COOKIE_COMMENT;
          break;
        }

      //  String constant "...."
      if (dwCookie & COOKIE_STRING)
        {
          if (pszChars[I] == '"')
            {
              dwCookie &= ~COOKIE_STRING;
              bRedefineBlock = TRUE;
            }
          continue;
        }

      if (pszChars[I] == '\'')
        {
          DEFINE_BLOCK (I, COLORINDEX_COMMENT);
          dwCookie |= COOKIE_COMMENT;
          break;
        }

      //  Normal text
      if (pszChars[I] == '"')
        {
          DEFINE_BLOCK (I, COLORINDEX_STRING);
          dwCookie |= COOKIE_STRING;
          continue;
        }

      if (bFirstChar)
        {
          if (!xisspace (pszChars[I]))
            bFirstChar = FALSE;
        }

      if (pBuf == NULL)
        continue;               //  We don't need to extract keywords,
      //  for faster parsing skip the rest of loop

      if (xisalnum (pszChars[I]) || pszChars[I] == '.' && I > 0 && (!xisalpha (pszChars[nPrevI]) && !xisalpha (pszChars[I + 1])))
        {
          if (nIdentBegin == -1)
            nIdentBegin = I;
        }
      else
        {
          if (nIdentBegin >= 0)
            {
              if (LangVB::IsVBKeyword (pszChars + nIdentBegin, I - nIdentBegin))
                {
                  DEFINE_BLOCK (nIdentBegin, COLORINDEX_KEYWORD);
                }
              else if (LangVB::IsVBNumber (pszChars + nIdentBegin, I - nIdentBegin))
                {
                  DEFINE_BLOCK (nIdentBegin, COLORINDEX_NUMBER);
                }
              else
                {
                  bool bFunction = FALSE;

                  for (int j = I; j < nLength; j++)
                    {
                      if (!xisspace (pszChars[j]))
                        {
                          if (pszChars[j] == '(')
                            {
                              bFunction = TRUE;
                            }
                          break;
                        }
                    }
                  if (bFunction)
                    {
                      DEFINE_BLOCK (nIdentBegin, COLORINDEX_FUNCNAME);
                    }
                }
              bRedefineBlock = TRUE;
              bDecIndex = TRUE;
              nIdentBegin = -1;
            }
        }
    }

  if (nIdentBegin >= 0)
    {
      if (LangVB::IsVBKeyword (pszChars + nIdentBegin, I - nIdentBegin))
        {
          DEFINE_BLOCK (nIdentBegin, COLORINDEX_KEYWORD);
        }
      else if (LangVB::IsVBNumber (pszChars + nIdentBegin, I - nIdentBegin))
        {
          DEFINE_BLOCK (nIdentBegin, COLORINDEX_NUMBER);
        }
      else
        {
          bool bFunction = FALSE;

          for (int j = I; j < nLength; j++)
            {
              if (!xisspace (pszChars[j]))
                {
                  if (pszChars[j] == '(')
                    {
                      bFunction = TRUE;
                    }
                  break;
                }
            }
          if (bFunction)
            {
              DEFINE_BLOCK (nIdentBegin, COLORINDEX_FUNCNAME);
            }
        }
    }

  dwCookie &= COOKIE_EXT_COMMENT;
  return dwCookie;
}
