/**
 *  @file PathContext.h
 *
 *  @brief Declarations of PathInfo and PathContext
 */
// RCS ID line follows -- this is updated by CVS
// $Id: PathContext.h 3998 2006-12-30 17:51:34Z kimmov $


#ifndef _PATH_CONTEXT_H_
#define _PATH_CONTEXT_H_

/**
 * @brief Information for one path.
 *
 * Path is stored in normalized format (no trailing slash).
 */
class PathInfo
{
public:
	PathInfo() {}
	PathInfo(const PathInfo &pi);

	CString GetPath(bool bNormalized = true) const;
	void SetPath(LPCTSTR path);
	void NormalizePath();

private:
	CString m_sPath;  /**< Directory / file path */
};

/**
 * @brief Holds path information of compared files/directories.
 */
class PathContext
{
public:
	PathContext();
	PathContext(LPCTSTR sLeft, LPCTSTR sRight);
	CString GetLeft(bool bNormalized = true) const;
	CString GetRight(bool bNormalized = true) const;
	CString GetPath(int index, bool bNormalized = true) const;
	void SetLeft(LPCTSTR path);
	void SetRight(LPCTSTR path);
	void SetPath(int index, LPCTSTR path);

private:
	PathInfo m_pathLeft; /**< First path (left path at start) */
	PathInfo m_pathRight; /**< Second path (right path at start */
};

/**
 * @brief Temp files for compared files
 */
class TempFileContext : public PathContext
{
public:
	~TempFileContext();
	bool CreateFiles(const PathContext &paths);
	bool FilesExist() const;
	void DeleteFiles();
	const CString & GetTempPath() const { return m_sTempPath; }

private:
	CString m_sTempPath;
};

#endif  // _PATH_CONTEXT_H_
