/**
 *  @file PluginManager.cpp
 *
 *  @brief Implementation of PluginManager class
 */ 
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PluginManager.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif



PluginManager::~PluginManager()
{
	// Manually free all items in cache
	// This could also be done with a DestructElements helper
	CString filteredFilenames;
	PluginFileInfo * fi=0;
	for (POSITION pos = m_pluginSettings.GetStartPosition(); pos; )
	{
		m_pluginSettings.GetNextAssoc(pos, filteredFilenames, fi);
		delete fi;
	}
	m_pluginSettings.RemoveAll();
}

/**
 * @brief retrieve relevant plugin settings for specified comparison
 */
void PluginManager::FetchPluginInfos(const CString& filteredFilenames, 
                                     PackingInfo ** infoUnpacker, 
                                     PrediffingInfo ** infoPrediffer, bool bSearchOnly)
{
	PluginFileInfo * fi = NULL;
	if (!m_pluginSettings.Lookup(filteredFilenames, fi) && !bSearchOnly)
	{
		fi = new PluginFileInfo();
		// This might be a good place to set any user-specified default values
		m_pluginSettings.SetAt(filteredFilenames, fi);
	}
  if (fi)
  {
    if (infoUnpacker && !*infoUnpacker)
	    *infoUnpacker = &fi->m_infoUnpacker;
    if (infoPrediffer && !*infoPrediffer)
	    *infoPrediffer = &fi->m_infoPrediffer;
  }
}

/**
 * @brief Store specified prediff choice for specified comparison
 */
void PluginManager::SetPrediffSetting(const CString & filteredFilenames, int newsetting)
{
	//PackingInfo * infoUnpacker = 0;
	PrediffingInfo * infoPrediffer = NULL;
	FetchPluginInfos(filteredFilenames, NULL, &infoPrediffer);
	infoPrediffer->Initialize(newsetting);
}

void PluginManager::SetPrediffer(const CString & filteredFilenames, const CString & prediffer)
{
	//PackingInfo * infoUnpacker = 0;
	PrediffingInfo * infoPrediffer = NULL;
	FetchPluginInfos(filteredFilenames, NULL, &infoPrediffer);
	infoPrediffer->pluginName = prediffer;
}