/** 
 * @file  PropCompare.cpp
 *
 * @brief Implementation of CPropCompare propertysheet
 */
// ID line follows -- this is updated by SVN
// $Id: PropCompare.cpp 4098 2007-02-03 15:57:56Z kimmov $

#include "stdafx.h"
#include "merge.h"
#include "PropCompare.h"
#include "OptionsDef.h"
#include "OptionsMgr.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CPropCompare property page

/** 
 * @brief Constructor.
 * @param [in] optionsMgr Pointer to COptionsMgr.
 */
CPropCompare::CPropCompare(COptionsMgr *optionsMgr) : CPropertyPage(CPropCompare::IDD)
 , m_pOptionsMgr(optionsMgr)
 , m_compareMethod(-1)
 , m_bIgnoreCase(false)
 , m_bIgnoreBlankLines(false)
 , m_bIgnoreEol(true)
 , m_nIgnoreWhite(-1)
 , m_bMovedBlocks(false)
 , m_bMatchSimilarLines(false)
 , m_bStopAfterFirst(false)
 , m_bFilterCommentsLines(false)
{
}

void CPropCompare::DoDataExchange(CDataExchange* pDX)
{
	CPropertyPage::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CPropCompare)
	DDX_CBIndex(pDX, IDC_COMPAREMETHODCOMBO, m_compareMethod);
  BOOL i = m_bIgnoreCase;
	DDX_Check(pDX, IDC_IGNCASE_CHECK, i);
  m_bIgnoreCase = i;
  i = m_bIgnoreBlankLines;
	DDX_Check(pDX, IDC_IGNBLANKS_CHECK, i);
  m_bIgnoreBlankLines = i;
  i = m_bFilterCommentsLines;
	DDX_Check(pDX, IDC_FILTERCOMMENTS_CHECK, i);
  m_bFilterCommentsLines = i;
  i = m_bIgnoreEol;
	DDX_Check(pDX, IDC_EOL_SENSITIVE, i);
  m_bIgnoreEol = i;
	DDX_Radio(pDX, IDC_WHITESPACE, m_nIgnoreWhite);
  i = m_bMovedBlocks;
	DDX_Check(pDX, IDC_MOVED_BLOCKS, i);
  m_bMovedBlocks = i;
  i = m_bMatchSimilarLines;
	DDX_Check(pDX, IDC_MATCH_SIMILAR_LINES, i);
  m_bMatchSimilarLines = i;
  i = m_bStopAfterFirst;
	DDX_Check(pDX, IDC_COMPARE_STOPFIRST, i);
  m_bStopAfterFirst = i;
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CPropCompare, CPropertyPage)
	//{{AFX_MSG_MAP(CPropCompare)
	ON_BN_CLICKED(IDC_COMPARE_DEFAULTS, OnDefaults)
	//}}AFX_MSG_MAP
	ON_CBN_SELCHANGE(IDC_COMPAREMETHODCOMBO, OnCbnSelchangeComparemethodcombo)
END_MESSAGE_MAP()

/** 
 * @brief Reads options values from storage to UI.
 * Property sheet calls this before displaying GUI to load values
 * into members.
 */
void CPropCompare::ReadOptions()
{
	m_nIgnoreWhite = m_pOptionsMgr->GetInt(OPT_CMP_IGNORE_WHITESPACE);
	m_bIgnoreBlankLines = m_pOptionsMgr->GetBool(OPT_CMP_IGNORE_BLANKLINES);
	m_bFilterCommentsLines = m_pOptionsMgr->GetBool(OPT_CMP_FILTER_COMMENTLINES);
	m_bIgnoreCase = m_pOptionsMgr->GetBool(OPT_CMP_IGNORE_CASE);
	m_bIgnoreEol = m_pOptionsMgr->GetBool(OPT_CMP_IGNORE_EOL);
	m_bMovedBlocks = m_pOptionsMgr->GetBool(OPT_CMP_MOVED_BLOCKS);
  m_bMatchSimilarLines = m_pOptionsMgr->GetBool(OPT_CMP_MATCH_SIMILAR_LINES);
	m_compareMethod = m_pOptionsMgr->GetInt(OPT_CMP_METHOD);
	m_bStopAfterFirst = m_pOptionsMgr->GetBool(OPT_CMP_STOP_AFTER_FIRST);
}

/** 
 * @brief Writes options values from UI to storage.
 * Property sheet calls this after dialog is closed with OK button to
 * store values in member variables.
 */
void CPropCompare::WriteOptions()
{
	m_pOptionsMgr->SaveOption(OPT_CMP_IGNORE_WHITESPACE, m_nIgnoreWhite);
	m_pOptionsMgr->SaveOption(OPT_CMP_IGNORE_BLANKLINES, m_bIgnoreBlankLines);
	m_pOptionsMgr->SaveOption(OPT_CMP_FILTER_COMMENTLINES, m_bFilterCommentsLines);
	m_pOptionsMgr->SaveOption(OPT_CMP_IGNORE_EOL, m_bIgnoreEol);
	m_pOptionsMgr->SaveOption(OPT_CMP_IGNORE_CASE, m_bIgnoreCase);
	m_pOptionsMgr->SaveOption(OPT_CMP_METHOD, m_compareMethod);
	m_pOptionsMgr->SaveOption(OPT_CMP_MOVED_BLOCKS, m_bMovedBlocks);
  m_pOptionsMgr->SaveOption(OPT_CMP_MATCH_SIMILAR_LINES, m_bMatchSimilarLines);
	m_pOptionsMgr->SaveOption(OPT_CMP_STOP_AFTER_FIRST, m_bStopAfterFirst);
}

/////////////////////////////////////////////////////////////////////////////
// CPropCompare message handlers

/** 
 * @brief Called before propertysheet is drawn.
 */
BOOL CPropCompare::OnInitDialog() 
{
	CPropertyPage::OnInitDialog();
	CComboBox * combo = (CComboBox*) GetDlgItem(IDC_COMPAREMETHODCOMBO);

	CString item = LoadResString(IDS_COMPMETHOD_FULL_CONTENTS);
	combo->AddString(item);
	item = LoadResString(IDS_COMPMETHOD_QUICK_CONTENTS);
	combo->AddString(item);
	item = LoadResString(IDS_COMPMETHOD_MODDATE);
	combo->AddString(item);
	item = LoadResString(IDS_COMPMETHOD_DATESIZE);
	combo->AddString(item);
	item = LoadResString(IDS_COMPMETHOD_SIZE);
	combo->AddString(item);
	combo->SetCurSel(m_compareMethod);

	CButton * pBtn = (CButton*) GetDlgItem(IDC_COMPARE_STOPFIRST);
	pBtn->EnableWindow(m_compareMethod == 1);
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

/** 
 * @brief Sets options to defaults
 */
void CPropCompare::OnDefaults()
{
	DWORD tmp;
	m_pOptionsMgr->GetDefault(OPT_CMP_METHOD, tmp);
	m_compareMethod = tmp;
	m_pOptionsMgr->GetDefault(OPT_CMP_IGNORE_WHITESPACE, tmp);
	m_nIgnoreWhite = tmp;
	m_pOptionsMgr->GetDefault(OPT_CMP_IGNORE_EOL, tmp);
	m_bIgnoreEol = tmp;
	m_pOptionsMgr->GetDefault(OPT_CMP_IGNORE_BLANKLINES, tmp);
	m_bIgnoreBlankLines = tmp;
	m_pOptionsMgr->GetDefault(OPT_CMP_FILTER_COMMENTLINES, tmp);
	m_bFilterCommentsLines = tmp;
	m_pOptionsMgr->GetDefault(OPT_CMP_IGNORE_CASE, tmp);
	m_bIgnoreCase = tmp;
	m_pOptionsMgr->GetDefault(OPT_CMP_MOVED_BLOCKS, tmp);
	m_bMovedBlocks = tmp;
  m_pOptionsMgr->GetDefault(OPT_CMP_MATCH_SIMILAR_LINES, tmp);
	m_bMatchSimilarLines = tmp;
	m_pOptionsMgr->GetDefault(OPT_CMP_STOP_AFTER_FIRST, tmp);
	m_bStopAfterFirst = tmp;
	UpdateData(FALSE);
}

/** 
 * @brief Called when compare method dropdown selection is changed.
 * Enables / disables "Stop compare after first difference" checkbox.
 * That checkbox is valid only for quick contents compare method.
 */
void CPropCompare::OnCbnSelchangeComparemethodcombo()
{
	CComboBox * pCombo = (CComboBox*) GetDlgItem(IDC_COMPAREMETHODCOMBO);
	CButton * pBtn = (CButton*) GetDlgItem(IDC_COMPARE_STOPFIRST);
	pBtn->EnableWindow(pCombo->GetCurSel() == 1);
}
