/** 
 * @file  paths.h
 *
 * @brief Declaration file for path routines
 */
// RCS ID line follows -- this is updated by CVS
// $Id: paths.h 4377 2007-07-22 13:53:19Z kimmov $

#ifndef paths_h_included
#define paths_h_included

class PathHelper
{
public:
  /**
   * @brief Possible values when checking for file/folder existence.
   */
  enum PATH_EXISTENCE
  {
	  DOES_NOT_EXIST, /**< File or folder does not exist. */
	  IS_EXISTING_FILE, /**< It is existing file */
	  IS_EXISTING_DIR, /**< It is existing folder */
  };

  static bool paths_EndsWithSlash(const CString & s);
  static PATH_EXISTENCE paths_DoesPathExist(LPCTSTR szPath);
  static void paths_normalize(CString & sPath);
  static CString paths_GetLongPath(const CString & sPath, bool bExpandEnvs = true);
  static bool paths_CreateIfNeeded(const CString & sPath);
  static LPCTSTR paths_GetTempPath(int * pnerr=NULL);
  static CString paths_GetTempFileName(LPCTSTR lpPathName, LPCTSTR lpPrefixString, int * pnerr=NULL);
  static PATH_EXISTENCE GetPairComparability(LPCTSTR pszLeft, LPCTSTR pszRight);
  static CString ExpandShortcut(const CString &inFile);
  static CString paths_ConcatPath(const CString & path, const CString & subpath);
  static CString paths_GetParentPath(const CString &path);
  static CString paths_GetLastSubdir(const CString &path);
  static BOOL paths_IsPathAbsolute(const CString &path);
  static CString paths_EnsurePathExist(const CString & sPath);
  static CString paths_GetFolderPath(HWND hwndOwner = NULL, int nFolder = CSIDL_APPDATA);
  static CString paths_GetWindowsDirectory();
  static CString paths_GetSystemDirectory();
  static CString paths_GetMyDocuments(HWND hWindow = NULL);
  static CString paths_GetApplicationDataFolder(HWND hWindow = NULL);
};
#endif // paths_h_included
