/** 
 * @file  stringdiffs.h
 *
 * @brief Interface file declaring sd_ComputeWordDiffs (q.v.)
 *
 */
// RCS ID line follows -- this is updated by CVS
// $Id: stringdiffs.h 2953 2006-01-17 04:31:41Z elsapo $

#ifndef stringdiffs_h_included
#define stringdiffs_h_included

/** @brief One difference between two strings */
struct wdiff {
	int start[2]; // 0-based, eg, start[0] is from str1
	int end[2]; // 0-based, eg, end[1] is from str2
	wdiff(int s1=0, int e1=0, int s2=0, int e2=0)
	{
		if (s1>e1) e1=s1-1;
		if (s2>e2) e2=s2-1;
		start[0] = s1;
		start[1] = s2;
		end[0] = e1;
		end[1] = e2;
	}
	wdiff(const wdiff & src)
	{
    start[0] = src.start[0];
		start[1] = src.start[1];
		end[0] = src.end[0];
		end[1] = src.end[1];
	}
  int GetLength(int index = 0)
  {
    if ((index < 0) || (index > 1))
      index = 0;
    int res = end[index] - start[index] + 1;
    if ((res < 0) || (start[index] < 0))
      res = 0;
    return res;
  }
};
typedef CArray<wdiff, wdiff&> wdiffarray; /**< An array of differences between two strings */


float sd_ComputeWordDiffs(const CString & str1, const CString & str2, wdiffarray * pDiffs,
                   bool case_sensitive = true, int whitespace = true, int breakType = 1, bool byte_level = false,
                   bool bComputeDiffStatus = true);

void sd_ComputeByteDiff(CString & str1, CString & str2, 
			bool casitive, int xwhite, 
			int &begin1, int &begin2, int &end1, int &end2);

#endif // stringdiffs_h_included
